@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2016 Sergey Poznyakoff
+   Copyright (C) 2016-2019 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -32,13 +32,13 @@ static size_t comp_count;
 
 static pies_depmap_t depmap;
 
-static int
+static inline int
 next_index (void)
 {
   return (cur + 1) % ARRAY_SIZE (comp_list);
 }
 
-static int
+static inline int
 prev_index (void)
 {
   return (cur + ARRAY_SIZE (comp_list) - 1) % ARRAY_SIZE (comp_list);
@@ -50,7 +50,7 @@ component_link (struct component *comp, struct component *ref)
   if (!ref)
     {
       struct complist *list = &comp_list[comp->listidx];
-  
+
       comp->prev = NULL;
       comp->next = list->head;
       if (list->head)
@@ -68,7 +68,7 @@ component_link (struct component *comp, struct component *ref)
 
       comp->prev = ref;
       comp->next = ref->next;
-      
+
       if ((x = ref->next))
 	x->prev = comp;
       else
@@ -84,12 +84,28 @@ component_append (struct component *comp)
   component_link (comp, comp_list[comp->listidx].tail);
 }
 
+void
+comp_array_remove (size_t i)
+{
+  struct component *comp = comp_array[i];
+
+  depmap_remove (depmap, i);
+  while (i < comp_count -1)
+    {
+      comp_array[i] = comp_array[i+1];
+      comp_array[i]->arridx = i;
+      i++;
+    }
+  component_free (comp);
+  comp_count--;
+}
+
 void
 component_unlink (struct component *comp)
 {
   struct complist *list = &comp_list[comp->listidx];
   struct component *x;
-  
+
   if ((x = comp->prev))
     x->next = comp->next;
   else
@@ -98,7 +114,7 @@ component_unlink (struct component *comp)
     x->prev = comp->prev;
   else
     list->tail = comp->prev;
-}  
+}
 
 int
 component_list_is_empty (void)
@@ -125,11 +141,11 @@ component_lookup_index (const char *tag)
   size_t i;
 
   for (i = 0; i < comp_count; i++)
-    if (comp_array[i] && strcmp (comp_array[i]->tag, tag) == 0)
+    if (strcmp (comp_array[i]->tag, tag) == 0)
       return i;
   return -1;
 }
-  
+
 struct component *
 component_create (const char *name)
 {
@@ -155,6 +171,7 @@ component_free (struct component *comp)
   component_unlink (comp);
   free (comp->tag);
   free (comp->program);
+  free (comp->command);
   if (comp->argv)
     {
       for (i = 0; i < comp->argc; i++)
@@ -201,17 +218,22 @@ component_ref_decr (struct component *comp)
 {
   assert (comp->ref_count > 0);
   if (--comp->ref_count == 0)
-    component_free (comp);
+    {
+      if (component_is_active (comp))
+	comp_array_remove (comp->arridx);
+      else
+	component_free (comp);
+    }
 }
 
 static int
 argvcmp (char **a, char **b)
 {
   size_t i;
-  
+
   if (!a != !b)
     return 1;
-  
+
   for (i = 0; a[i]; i++)
     if (!b[i] || strcmp (b[i], a[i]))
       return 1;
@@ -250,7 +272,7 @@ redirector_cmp (struct redirector const *a, struct redirector const *b)
 
   return 0;
 }
-      
+
 static int
 component_match (struct component *comp, struct component *ref)
 {
@@ -258,7 +280,7 @@ component_match (struct component *comp, struct component *ref)
 #define EQ(memb) MATCH (comp->memb != ref->memb)
 #define FN(memb,fun) MATCH (fun (comp->memb, ref->memb))
 #define FNP(memb,fun) MATCH (fun (&comp->memb, &ref->memb))
-  
+
   EQ (mode);
   FN (tag, safe_strcmp);
   FN (program, safe_strcmp);
@@ -299,7 +321,7 @@ complist_find_match (int idx, struct component *ref)
 {
   struct complist *list = &comp_list[idx];
   struct component *comp;
-  
+
   for (comp = list->head; comp && component_match (comp, ref);
        comp = comp->next)
     ;
@@ -331,7 +353,7 @@ component_merge (struct component *comp, struct component *ref)
   pies_acl_free (comp->list_acl);
   comp->list_acl = ref->list_acl;
   ref->list_acl = NULL;
-  
+
   pies_acl_free (comp->adm_acl);
   comp->adm_acl = ref->adm_acl;
   ref->adm_acl = NULL;
@@ -366,7 +388,7 @@ prog_is_leftover (struct prog *prog)
 }
 
 /* If PROG is a leftover, mark it for termination.  If it is a listener,
-   terminate it immediately.  This ensures that all decomissioned sockets
+   terminate it immediately.  This ensures that all decommissioned sockets
    are closed before the subsequent call to progman_create_sockets, which
    might need to reopen some of them.
 */
@@ -388,32 +410,51 @@ list_str_cmp (const void *a, const void *b)
   return safe_strcmp (a, b);
 }
 
+/* Report cyclic dependency starting at IDX.  Mark each element with
+   CF_REMOVE for subsequent removal.
+   DP is a transitive closure of depmap.
+*/
 static void
-component_log_dep (size_t idx)
+report_cyclic_dependency (pies_depmap_t dp, size_t idx)
 {
-  pies_depmap_pos_t pos;
-  size_t n;
-  
-  logmsg_printf (LOG_NOTICE, "%s -> ", comp_array[idx]->tag);
-  for (n = depmap_first (depmap, depmap_col, idx, &pos);
-       n != (size_t)-1;
-       n = depmap_next (depmap, pos))
+  size_t i;
+
+  i = idx;
+  do
     {
-      logmsg_printf (LOG_NOTICE, "%s -> ", comp_array[n]->tag);
+      size_t n;
+      pies_depmap_pos_t pos;
+
+      logmsg_printf (LOG_NOTICE, "%s -> ", comp_array[i]->tag);
+      comp_array[i]->flags |= CF_REMOVE;
+      for (n = depmap_first (depmap, depmap_col, i, &pos);
+	   n != (size_t)-1;
+	   n = depmap_next (depmap, pos))
+	{
+	  if (n == i)
+	    continue;
+	  if (depmap_isset (dp, n, i) && depmap_isset (dp, n, n))
+	    break;
+	}
+      depmap_end (pos);
+
+      if (n == (size_t)-1)
+	break;
+      i = n;
     }
-  depmap_end (pos);
+  while (i != idx);
   logmsg_printf (LOG_NOTICE, "%s\n", comp_array[idx]->tag);
-}  
+}
 
 void
 component_build_depmap (void)
 {
   size_t i;
   pies_depmap_t dp;
-  
+
   free (depmap);
   depmap = depmap_alloc (comp_count);
-  for (i = 0; i < comp_count; i++)
+  for (i = 0; i < comp_count; )
     {
       struct component *comp = comp_array[i];
       struct grecs_list_entry *ep;
@@ -429,15 +470,12 @@ component_build_depmap (void)
 			_("component %s depends on %s, "
 			  "which is not declared"),
 			comp->tag, tag);
-		component_free (comp);
-		comp_array[i] = NULL;
-		depmap_clear_all (depmap, depmap_row, i);
-		depmap_clear_all (depmap, depmap_col, i);
+		comp_array_remove (i);
 		continue;
 	      }
 	    depmap_set (depmap, i, tgt);
 	  }
-      
+
       if (comp->depend)
 	for (ep = comp->depend->head; ep; ep = ep->next)
 	  {
@@ -452,22 +490,27 @@ component_build_depmap (void)
 	      }
 	    depmap_set (depmap, tgt, i);
 	  }
+
+      i++;
     }
 
   dp = depmap_copy (depmap);
   depmap_tc (dp);
   for (i = 0; i < comp_count; i++)
-    if (depmap_isset (dp, i, i))
+    if (!(comp_array[i]->flags & CF_REMOVE) && depmap_isset (dp, i, i))
       {
 	logmsg (LOG_ERR, _("component %s depends on itself"),
 		comp_array[i]->tag);
-	component_log_dep (i);
-	component_free (comp_array[i]);
-	comp_array[i] = NULL;
-	depmap_clear_all (depmap, depmap_row, i);
-	depmap_clear_all (depmap, depmap_col, i);
-	continue;
+	report_cyclic_dependency (dp, i);
       }
+
+
+  for (i = 0; i < comp_count;)
+    if (comp_array[i]->flags & CF_REMOVE)
+      comp_array_remove (i);
+    else
+      i++;
+
   free (dp);
 }
 
@@ -492,7 +535,7 @@ component_config_commit (void)
     comp_array = grecs_realloc (comp_array, i * sizeof (comp_array[0]));
   comp_count = i;
 
-  /* Rearrange components, registering prog entries for the new ones */
+  /* Rearrange components, registering entries for the new ones */
   for (comp = list->head, i = 0; comp; comp = comp->next, i++)
     {
       match = complist_find_match (prev, comp);
@@ -516,7 +559,7 @@ component_config_commit (void)
       progman_foreach (mark_prog, NULL);
       pies_schedule_children (PIES_CHLD_GC);
     }
-  
+
   /* Build dependency map */
   component_build_depmap ();
 
@@ -567,7 +610,7 @@ component_verify (struct component *comp, grecs_locus_t *locus)
   else if (!comp->argv)
     COMPERR (grecs_error,
 	     "%s", _("missing command line"));
-  
+
   if (ISCF_TCPMUX (comp->flags))
     {
       comp->mode = pies_comp_inetd;
@@ -580,7 +623,7 @@ component_verify (struct component *comp, grecs_locus_t *locus)
 	COMPERR (grecs_error,
 		 "%s", _("\"internal\" used without \"service\""));
     }
-  
+
   if (comp->pass_fd_socket && comp->mode != pies_comp_pass_fd)
     COMPERR (grecs_error,
 	     "%s", _("pass-fd-socket ignored: wrong mode"));
@@ -591,7 +634,7 @@ component_verify (struct component *comp, grecs_locus_t *locus)
 	COMPERR (grecs_error,
 		 "%s", _("socket ignored: wrong mode"));
       break;
-      
+
     case pies_comp_pass_fd:
       if (!comp->pass_fd_socket)
 	COMPERR (grecs_error,
@@ -610,7 +653,7 @@ component_verify (struct component *comp, grecs_locus_t *locus)
 			     "file name or chdir must be specified"));
 	}
       /* Fall through */
-      
+
     case pies_comp_accept:
       if (!comp->socket_url)
 	{
@@ -619,7 +662,7 @@ component_verify (struct component *comp, grecs_locus_t *locus)
 	  return 1;
 	}
       break;
-      
+
     case pies_comp_inetd:
       if (ISCF_TCPMUX (comp->flags))
 	{
@@ -678,7 +721,7 @@ component_verify (struct component *comp, grecs_locus_t *locus)
     default:
       break;
     }
-  
+
   for (i = RETR_OUT; i <= RETR_ERR; i++)
     {
       if (comp->redir[i].type == redir_file
@@ -697,7 +740,7 @@ component_verify (struct component *comp, grecs_locus_t *locus)
 		     comp->redir[i].v.file);
 	}
     }
-  
+
   return header;
 #undef COMPERR
 }
@@ -705,13 +748,36 @@ component_verify (struct component *comp, grecs_locus_t *locus)
 void
 component_finish (struct component *comp, grecs_locus_t *locus)
 {
+  if (comp->flags & CF_SHELL)
+    {
+      comp->argc = 3;
+      comp->argv = grecs_calloc (comp->argc + 1, sizeof (comp->argv[0]));
+      comp->argv[0] = grecs_strdup (comp->program ? comp->program : "/bin/sh");
+      comp->argv[1] = grecs_strdup ("-c");
+      comp->argv[2] = grecs_strdup (comp->command);
+      comp->argv[3] = NULL;
+    }
+  else
+    {
+      struct wordsplit ws;
+      if (wordsplit (comp->command, &ws, WRDSF_DEFFLAGS))
+	{
+	  grecs_error (locus, 0, "wordsplit: %s",
+		       wordsplit_strerror (&ws));
+	  component_free (comp);
+	  return;
+	}
+      wordsplit_get_words (&ws, &comp->argc, &comp->argv);
+      wordsplit_free (&ws);
+    }
+
   if (comp->prereq)
     comp->prereq->cmp = list_str_cmp;
   if (comp->depend)
     comp->depend->cmp = list_str_cmp;
   if (comp->privs.groups)
     comp->privs.groups->cmp = list_str_cmp;
-  
+
   if (component_verify (comp, locus))
     {
       component_free (comp);
@@ -748,7 +814,7 @@ component_get (size_t n)
 }
 
 void
-components_dump_depmap (void)
+depmap_dump (pies_depmap_t dpm)
 {
   size_t i, j;
 
@@ -759,9 +825,9 @@ components_dump_depmap (void)
   printf ("\n");
   for (i = 0; i < comp_count; i++)
     {
-      printf ("%2lu ", (unsigned long)i);
+      printf ("%2lu", (unsigned long)i);
       for (j = 0; j < comp_count; j++)
-	printf (" %c ", depmap_isset (depmap, i, j) ? 'X' : ' ');
+	printf ("  %c", depmap_isset (dpm, i, j) ? 'X' : ' ');
       printf ("\n");
     }
   printf ("\n%s:\n", _("Legend"));
@@ -769,13 +835,19 @@ components_dump_depmap (void)
     printf ("%2lu: %s\n", (unsigned long)i, comp_array[i]->tag);
 }
 
+void
+components_dump_depmap (void)
+{
+  depmap_dump (depmap);
+}
+
 void
 component_trace (size_t idx, enum pies_depmap_direction dir)
 {
   pies_depmap_pos_t pos;
   size_t n;
   int delim = ':';
-    
+
   logmsg_printf (LOG_NOTICE, "%s", comp_array[idx]->tag);
   for (n = depmap_first (depmap, dir, idx, &pos);
        n != (size_t)-1;
@@ -844,8 +916,3 @@ component_foreach (int (*filter) (struct component *, void *), void *data)
     }
   return rc;
 }
-
-
-
-
-
