@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2008-2011, 2013-2016 Sergey Poznyakoff
+   Copyright (C) 2008-2019 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -104,7 +104,7 @@ struct action
   enum return_action act;  /* Action to take when the component terminates */
   char *addr;              /* Addresses to notify about it. */
   char *message;           /* Notification mail. */
-  char *command;           /* Execute this command */    
+  char *command;           /* Execute this command */
 };
 
 
@@ -137,14 +137,24 @@ enum pies_comp_mode
        component via the UNIX domain socket. Corresponds to
        `start_action = pass' in MeTA1. */
     pies_comp_pass_fd,
-    
+
+    /* Components of this type runs once on program startup. Running other
+       components is delayed until the last startup component finishes. */
+    pies_comp_startup,
+
+    /* Components of this type are run right before program termination.
+       They have shutdown_timeout seconds to finish their job and terminate
+       gracefully, othervise they will be terminated forcefully via SIGTERM
+       (and SIGKILL, for persisting ones). */
+    pies_comp_shutdown,
+
     /*
     ** Init-style components
     */
     pies_mark_sysvinit,
     /* Start the process when the specified runlevel is entered and wait
        for its termination */
-    pies_comp_wait = pies_mark_sysvinit,		
+    pies_comp_wait = pies_mark_sysvinit,
     /* Execute the component once, when the specified runlevel is entered */
     pies_comp_once,
     /* Execute the component during system boot. Ignore runlevel settings. */
@@ -162,7 +172,7 @@ enum pies_comp_mode
     pies_comp_powerokwait,
     /* Execute the process when SIGINT is delivered, i.e. someone has
        pressed the Ctrl+Alt+Del combination. */
-    pies_comp_ctrlaltdel,	
+    pies_comp_ctrlaltdel,
     /* Execute the component when a specified ondemand runlevel is called */
     pies_comp_ondemand,
     /* Execute the component on the system boot. */
@@ -184,20 +194,22 @@ enum pies_comp_mode
 
 #define CF_DISABLED    0x001 /* The componenet is disabled */
 #define CF_PRECIOUS    0x002 /* The component is precious (should not
-			        be disabled) */
+				be disabled) */
 #define CF_WAIT        0x004 /* Wait for the component instance to
-			        terminate. */
+				terminate. */
 #define CF_TCPMUX      0x008 /* A plain TCPMUX service */
 #define CF_TCPMUXPLUS  0x010 /* A TCPMUX-plus service, i.e. pies
-			        must emit a '+' response before starting
-			        it */
+				must emit a '+' response before starting
+				it */
 #define CF_INTERNAL    0x020 /* An internal inetd service */
 #define CF_SOCKENV     0x040 /* Component wants socket information in
-			        the environment */
+				the environment */
 #define CF_RESOLVE     0x080 /* Resolve IP addresses */
 #define CF_SIGGROUP    0x100 /* Send signals to the process group */
-
 #define CF_NULLINPUT   0x200 /* Provide null input stream */
+#define CF_SHELL       0x400 /* Invoke via sh -c */
+
+#define CF_REMOVE      0xf000 /* Marked for removal */
 
 #define ISCF_TCPMUX(f) ((f) & (CF_TCPMUX | CF_TCPMUXPLUS))
 
@@ -208,12 +220,13 @@ struct component
   struct component *prev, *next; /* Components form doubly-linked list. */
   int listidx;                   /* Index of the list. */
   size_t arridx;                 /* Index of this component. */
-  size_t ref_count;              /* Reference count. */ 
+  size_t ref_count;              /* Reference count. */
   struct prog *prog;             /* Prog associated with this component. */
-  
+
   enum pies_comp_mode mode;
   char *tag;               /* Entry tag (for diagnostics purposes) */
   char *program;           /* Program name */
+  char *command;           /* Full command line */
   size_t argc;             /* Number of command line arguments */
   char **argv;             /* Program command line */
   char **env;              /* Program environment */
@@ -230,7 +243,7 @@ struct component
 
   /* For exec (init) components */
   char *runlevels;
-  
+
   /* For inetd components */
   size_t max_rate;         /* Maximum number of invocations per minute */
   size_t max_ip_connections; /* Max. number of connections per IP address */
@@ -251,9 +264,9 @@ struct component
   char *access_denied_message;
   char *max_instances_message;
   char *max_ip_connections_message;
-  
+
   /* Redirectors: */
-  int facility;	         /* Syslog facility. */
+  int facility;		 /* Syslog facility. */
   struct redirector redir[2];   /* Repeaters for stdout and stderr */
   /* Actions to execute on various exit codes: */
   struct grecs_list *act_list;
@@ -332,6 +345,7 @@ int pies_read_config (void);
 int pies_reread_config (void);
 
 void register_prog (struct component *comp);
+void program_init_startup (void);
 int progman_waiting_p (void);
 void progman_start (void);
 void progman_gc (void);
@@ -374,9 +388,7 @@ pies_depmap_t depmap_copy (pies_depmap_t dpm);
 void depmap_set (pies_depmap_t dmap, size_t row, size_t col);
 int depmap_isset (pies_depmap_t dmap, size_t row, size_t col);
 void depmap_clear (pies_depmap_t dmap, size_t row, size_t col);
-
-void depmap_clear_all (pies_depmap_t dmap, enum pies_depmap_direction dir,
-		       size_t coord);
+void depmap_remove (pies_depmap_t dmap, size_t n);
 
 void depmap_tc (pies_depmap_t dmap);
 size_t depmap_first (pies_depmap_t dmap, enum pies_depmap_direction dir,
@@ -594,7 +606,7 @@ struct sysvinit_request
 #define SYSV_ACCT_BOOT 0
 #define SYSV_ACCT_RUNLEVEL 1
 #define SYSV_ACCT_PROC_START 2
-#define SYSV_ACCT_PROC_STOP 3	
+#define SYSV_ACCT_PROC_STOP 3
 
 void sysvinit_acct (int what, const char *user, const char *id, pid_t pid,
 		    const char *line);
