@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2015, 2016 Sergey Poznyakoff
+   Copyright (C) 2015-2019 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -656,7 +656,7 @@ shttp_request_send (struct shttp_connection *conn, int method, char const *uri)
   if (conn->b64auth)
     shttp_send_line (conn, "Authorization: Basic %s", conn->b64auth);
   
-  grecs_symtab_enumerate (conn->req.headers, send_header, conn);
+  grecs_symtab_foreach (conn->req.headers, send_header, conn);
   if (conn->req.content_length)
     {
       shttp_send_line (conn, "Content-Length: %lu", conn->req.content_length);
@@ -1052,7 +1052,9 @@ struct kwtrans mode_trans[] = {
   { "exec",         'C' },
   { "accept",       'A' }, 
   { "inetd",        'I' }, 
-  { "pass_fd",      'P' }, 
+  { "pass_fd",      'P' },
+  { "startup",      'S' },
+  { "shutdown",     'E' },
   { "wait",         'W' },  
   { "once",         'c' }, 
   { "boot",         'B' }, 
